from typing import *
import pytest

from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
def test__values():
    class Example(Base_AttrKit):
        A1: Any
        A2: Any = None
        A3 = None
        DICT: dict = {}

    try:
        Example()
        assert False
    except:
        assert True

    assert Example(a1=1).A1 == 1
    assert Example(1, a1=2).A1 == 2

    assert Example(1).A1 == 1
    assert Example(1).A2 == None
    assert Example(1).A3 == None

    assert Example(1, 1, 1).A1 == 1
    assert Example(1, 1, 1).A2 == 1
    assert Example(1, 1, 1).A3 == None
    assert Example(1, 1, a3=1).A3 == 1

    # mutable
    victim = Example(1, 1, a3=1)
    assert victim.DICT == Example.DICT
    assert victim.DICT is not Example.DICT

    victim.DICT[1]=1
    assert victim.DICT[1] == 1
    assert victim.DICT != Example.DICT


# ---------------------------------------------------------------------------------------------------------------------
def test__eq():
    class Example:
        A0: Any
        A1: Any = 1

    assert AttrKit_Blank(a1=1) != Example()
    assert AttrKit_Blank(A1=1) == Example()
    assert AttrKit_Blank(a1=11) != Example()
    assert AttrKit_Blank(a0=1) != Example()

    try:
        AttrKit_AuthTgBot(1)
        assert False
    except:
        assert True

    assert AttrKit_AuthTgBot(1, 2, 3).token == 3


# =====================================================================================================================
def test__cls_name():
    assert NestInit_AnnotsAttrByKwArgs().__class__.__name__ == f"NestInit_AnnotsAttrByKwArgs"

    class Victim(NestInit_AnnotsAttrByKwArgs):
        A1: Any = None

    assert Victim().__class__.__name__ == f"Victim"


# =====================================================================================================================



from typing import *
import time

from base_aux.testplans.devices import *
from base_aux.buses.m1_serial2_client_derivatives import *


# =====================================================================================================================
class Device(SerialClient_FirstFree_AnswerValid, Base_Device):  # IMPORTANT! KEEP Serial FIRST Nesting!
    LOG_ENABLE = True
    RAISE_CONNECT = False
    BAUDRATE = 115200
    PREFIX = "ATC:03:"
    EOL__SEND = b"\n"

    REWRITEIF_READNOANSWER = 0
    REWRITEIF_NOVALID = 0

    NAME = "ATC"
    DESCRIPTION: str = "ATC for PSU"

    # def address__validate(self) -> bool:  # NO NEED!
    #     return  self.write_read__last_validate("get name", self.NAME, prefix=self.PREFIX)

    def __init__(self, index: int = None, **kwargs):    # FIXME: decide to delete this!!!
        """
        :param index: None is only for SINGLE!
        """
        if index is not None:
            self.INDEX = index
        super().__init__(**kwargs)

    def dev__load_info(self) -> None:
        if not self.SN:
            self.SN = self.write_read__last("get SN")
            self.FW = self.write_read__last("get FW")
            self.MODEL = self.write_read__last("get MODEL")

    def connect__validate(self) -> bool:
        result = (
            self.address_check__resolved()  # fixme: is it really need here???
        )
        if result:
            self.dev__load_info()

        return result

    # DETECT --------------------------------
    @property
    def DEV_FOUND(self) -> bool:
        return self.address_check__resolved()


# =====================================================================================================================
class DeviceDummy(SerialClient_FirstFree_AnswerValid, Base_Device):  # IMPORTANT! KEEP Serial FIRST Nesting!
    @property
    def DEV_FOUND(self) -> bool:
        return True

    def address__validate(self) -> bool:
        return True

    def connect__validate(self) -> bool:
        return True

    def connect(self, *args, **kwargs) -> bool:
        return True


# =====================================================================================================================
if __name__ == "__main__":
    # emu = Atc_Emulator()
    # emu.start()
    # emu.wait()

    dev = Device()
    print(f"=======before {dev.ADDRESS=}")
    print(f"{dev.addresses_system__detect()=}")
    print(f"{dev.connect()=}")
    print(f"{dev.connect__only_if_address_resolved()=}")
    print(f"{dev.addresses_system__detect()=}")
    print(f"=======after {dev.ADDRESS=}")
    exit()

    print(f'{dev.SET(V12="ON")=}')
    print(f'{dev.SET(VIN=230)=}')
    print(f'{dev.SET(VOUT=220)=}')
    time.sleep(0.3)
    print(f'{dev.reset()=}')

    # print(f"{dev.address__validate()=}")
    # print(f"{dev.address__validate()=}")
    # print(f"{dev.address__validate()=}")
    #
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.write_read_line_last('get name')=}")
    # print(f"{dev.disconnect()=}")
    print(f"{dev.ADDRESS=}")


# =====================================================================================================================



from typing import *
from base_aux.testplans.devices import *
from base_aux.buses.m1_serial2_client_derivatives import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *


# =====================================================================================================================
class Device(SerialClient_FirstFree_AnswerValid, Base_Device):  # IMPORTANT! KEEP Serial FIRST Nesting!
    LOG_ENABLE = True
    RAISE_CONNECT = False
    BAUDRATE = 115200
    EOL__SEND = b"\n"

    REWRITEIF_READNOANSWER = 0
    REWRITEIF_NOVALID = 0

    # INFO --------------------------------
    NAME: str = "PTB"
    DESCRIPTION: str = "PTB for PSU"

    # @property
    # def DUT_SN(self) -> str:      # TODO: USE DIRECT FINDING!!!???
    #     return f"SN_{self.INDEX}"

    def __init__(self, index: int = None, **kwargs):    # FIXME: decide to delete this!!!
        """
        :param index: None is only for SINGLE!
        """
        if index is not None:
            self.INDEX = index
        super().__init__(**kwargs)

    # DETECT --------------------------------
    @property
    def DEV_FOUND(self) -> bool:
        return self.address_check__resolved()

    @property
    def PREFIX(self) -> str:
        return f"PTB:{self.INDEX+1:02d}:"

    # def address__validate(self) -> bool:  # NO NEED! only for manual!
    #     result = (
    #             self.write_read__last_validate("get name", self.NAME, prefix="*:")
    #             and
    #             self.write_read__last_validate("get addr", EqValid_NumParsedSingle_Success(self.INDEX+1), prefix="*:")
    #             # and
    #             # self.write_read__last_validate("get prsnt", "0")
    #     )
    #     if result:
    #         self.dev__load_info()
    #
    #     return result

    def dev__load_info(self) -> None:
        if not self.SN:
            self.SN = self.write_read__last("get SN")
            self.FW = self.write_read__last("get FW")
            self.MODEL = self.write_read__last("get MODEL")

            self.DUT_SN = self.write_read__last("get PSSN")
            self.DUT_FW = self.write_read__last("get PSFW")
            self.DUT_MODEL = self.write_read__last("get PSMODEL")

    def connect__validate(self) -> bool:
        result = (
            self.address_check__resolved()  # fixme: is it really need here???
            and
            self.write_read__last_validate("get prsnt", "1")
        )
        if result:
            self.dev__load_info()

        return result


# =====================================================================================================================
class DeviceDummy(SerialClient_FirstFree_AnswerValid, Base_Device):  # IMPORTANT! KEEP Serial FIRST Nesting!
    @property
    def DEV_FOUND(self) -> bool:
        return True

    def address__validate(self) -> bool:
        return True

    def connect__validate(self) -> bool:
        return True

    def connect(self, *args, **kwargs) -> bool:
        return True


# =====================================================================================================================
def _explore():
    pass

    # emu = Ptb_Emulator()
    # emu.start()
    # emu.wait()

    dev = Device(0)
    print(f"{dev.connect()=}")
    print(f"{dev.ADDRESS=}")
    print(f"{dev.address_check__resolved()=}")

    if not dev.address_check__resolved():
        return

    # dev.write_read__last("get sn")
    # dev.write_read__last("get fru")
    # dev.write_read__last("test sc12s")
    # dev.write_read__last("test ld12s")
    # dev.write_read__last("test gnd")
    # dev.write_read__last("test")
    # dev.write_read__last("get status")
    # dev.write_read__last("get vin")


if __name__ == "__main__":
    _explore()


# =====================================================================================================================



from base_aux.base_lambdas.m1_lambda2_derivatives import *
from base_aux.aux_dict.m3_dict_ga1_simple import *


# =====================================================================================================================
dict_example = {
    "lowercase": "lowercase",
    # "nested": {"n1":1},
}


class Victim(DictGaAnnotRequired):
    lowercase: str


# =====================================================================================================================
def test__obj():
    # victim = DictGaAnnotRequired()
    # assert victim == {}
    #
    # victim = DictGaAnnotRequired(hello=1)
    # assert victim == {"hello": 1}

    try:
        victim = Victim()
    except:
        assert True
    else:
        assert False


def test__dict_only():
    assert Lambda_TrySuccess(DictGaAnnotRequired) == True
    assert Lambda_TrySuccess(DictGaAnnotRequired)

    assert Lambda_TryFail(DictGaAnnotRequired) != True
    assert not Lambda_TryFail(DictGaAnnotRequired)

    assert Lambda_TrySuccess(DictGaAnnotRequired, **dict_example)
    assert Lambda_TrySuccess(DictGaAnnotRequired, lowercase="lowercase")
    assert Lambda_TrySuccess(DictGaAnnotRequired, LOWERCASE="lowercase")


def test__with_annots():
    assert Lambda_TryFail(Victim)
    assert not Lambda_TrySuccess(Victim)

    victim = Victim(lowercase="lowercase")
    assert victim["lowercase"] == "lowercase"

    assert Lambda_TrySuccess(Victim, **dict_example)
    assert Lambda_TrySuccess(Victim, lowercase="lowercase")
    assert Lambda_TrySuccess(Victim, LOWERCASE="lowercase")

    assert Lambda_TryFail(Victim, hello="lowercase")

    victim = Victim(lowercase="lowercase")
    assert victim == {"lowercase": "lowercase"}
    assert victim[1] == None
    assert victim.lowercase == "lowercase"


# =====================================================================================================================



from typing import *
import pytest

from base_aux.aux_dict.m2_dict_ic import *


# =====================================================================================================================
def test__dict_ic():
    victim1 = DictIcKeys()
    victim1['NAme'] = 'VALUE'
    victim1[1] = 1

    victim2 = DictIcKeys(NAme='VALUE')
    victim2[1] = 1

    victim3 = DictIcKeys({"NAme": 'VALUE', 1: 1})

    for victim in [victim1, victim2, victim3, ]:
        assert len(victim) == 2

        assert list(victim) == ["NAme", 1]

        # EQ
        assert victim == {1: 1, "naME": 'VALUE'}
        assert victim == {"naME": 'VALUE', 1: 1}

        assert victim != {"naME": 'VALUE', 1: 11}
        assert victim != {"naME": 'VALUE', 11: 1}
        assert victim != {"naME": 'VALUE2', 1: 1}
        assert victim != {"naME2": 'VALUE', 1: 1}

        # CONTAIN
        assert 'naME' in victim
        assert 1 in victim
        assert 0 not in victim

        # ACCESS
        Lambda(lambda: victim[0]).expect__check_assert(Exception)   # keep original behaviour - maybe need switch to None???
        Lambda(lambda: victim.get(0)).expect__check_assert(None)

        assert victim[1] == 1
        assert victim.get(1) == 1
        assert victim['name'] == "VALUE"
        assert victim['NAME'] == "VALUE"
        assert victim.get('NAME') == "VALUE"

        # update
        len0 = len(victim)
        victim['name'] = 'VALUE2'
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE2"

        victim['NAME'] = 'VALUE3'
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE3"

        victim.update({'NaMe': 'VALUE4'})
        assert len(victim) == len0
        assert victim['name'] == victim['NAME'] == "VALUE4"

        assert list(victim) == ["NAme", 1]

        # del ------
        try:
            del victim['name222']
        except:
            assert True
        else:
            assert False

        del victim['name']
        assert list(victim) == [1, ]
        victim['NAme'] = 'VALUE'

        # pop ------
        try:
            victim.pop('name222')
        except:
            assert True
        else:
            assert False

        assert victim.pop('name') == "VALUE"
        assert list(victim) == [1, ]
        # victim['NAme'] = 'VALUE'


# =====================================================================================================================



DESCRIPTION: str = """
designed to apply testplan for several DUTs

## ПОНЯТИЯ
    TC - TestCase
    TP - TestPlan
    DUT - Device Under Test - тестируемое устройство

## АРХИТЕКТУРА
- тестплан
    - работает в потоке,
    - может быть остановлен в любой момент terminate(), при этом завершает все запущенные тесткейсы
    - имеет настройки которые принимаются всеми тесткейсами за базовые и могут быть перезаписаны ими для себя
    - имеет списки классов TC и обьектов DUT (генерирует обьекты TC для каждого DUT)
    - для себя не разделяет обьекты тесткейсов, работает строго с классом тесткейса,
    - выполняет все тесткейсы в порядке их следования на списке DUT
    - в один момент времени выполняется только один класс тесткейса
- тесткейсы
    - работают в потоке,
    - может быть остановлен в любой момент terminate(), при этом завершаются безопасно (исполняются все teardown обьектов и глобальный классовый тесткейса), 
    - представляет собой класс инициируемый с входным параметром DUT,
    - выполняются тесткейсы строго по очереди,
    - каждый тесткейс выполняется на всех устройствах либо асинхронно, либо синхронно в зависимости от настройки,
    - работа тесткейса полностью управляется классом тесткейса на серии устройств (возможно выполнение парных тестов с выбором нужных пар внутри тесткейса),
- настройки
    - управление
        - SKIP всех возможных вариантов (полностью тесткейс для всех устройств, полностью DUT для всех TC, отдельный TC на отдельном DUT),
        - выполнение тесткейса синхронно/асинхронно
    - данные для использования в тесткейсах
        - реализовано в файлах JSON
        - для каждого тесткейса и общие для тестплана (кейсовые накладываются на плановые)
- API 
    - минимальное API и запуск
"""

FEATURES = [
    # "feat1",
    # ["feat2", "block1", "block2"],

    ["[THREADS]",
        "safe work in independent TCs",
        "safe stop process at any moment by terminate",
     ],
    ["[SKIP]",
        "tc", "tc on dut", "dut"
     ],
    ["[DEV_COLUMN]",
        "keep all in one instance",
        "use variants: single device for all duts or list for pairing each dut",
     ],
]


# =====================================================================================================================



from typing import *
import pathlib

from base_aux.base_values.m2_value_special import VALUE_SPECIAL
from base_aux.base_types.m0_static_types import TYPES


# =====================================================================================================================
@final
class TYPING:
    """
    GOAL
    ----
    collect all typing USER variants
    """
    ELEMENTARY = Union[*TYPES.ELEMENTARY]

    # -----------------------------------------------------------------------------------------------------------------
    ARGS_FINAL = tuple[Any, ...]
    ARGS_DRAFT = Union[Any, ARGS_FINAL, 'ArgsKwargs']           # you can use direct single value

    KWARGS_FINAL = dict[str, Any]
    KWARGS_DRAFT = Union[None, KWARGS_FINAL, 'ArgsKwargs']  # if passed NONE - no data!

    # -----------------------------------------------------------------------------------------------------------------
    PATH_FINAL = pathlib.Path
    PATH_DRAFT = Union[str, PATH_FINAL]

    STR_FINAL = str
    STR_DRAFT = Union[STR_FINAL, Any]

    ATTR_FINAL = str
    ATTR_DRAFT = str | int | Any

    # -----------------------------------------------------------------------------------------------------------------
    DICT_ANY_NONE = dict[Any, None]             # just to show - dict with None values after clearing!
    DICT_ANY_ANY = dict[Any, Any]               # just to show - dict could be any! on keys/values
    DICT_STR_ANY = dict[str, Any]               # just to show - dict could be any! on values! not just an elementary1
    DICT_STR_ELEM = DICT_JSON_ANY = dict[str, ELEMENTARY]       # just to show - parsed by json - dict
    DICT_STR_STR = DICT_INI = dict[str, str]               # just to show - parsed by ini!
    JSON_ANY = ELEMENTARY | DICT_STR_ELEM  # just to show - parsed by json - any object

    DICT_ANY_TUPLE_ANY = dict[Any, tuple[Any, ...]]     # as DICT_DIFF_DICTS
    DICT_STR_TUPLE_ANY = dict[str, tuple[Any, ...]]     # as DICT_DIFF_ATTRS

    # -----------------------------------------------------------------------------------------------------------------
    ITERABLE_ORDERED = Union[dict, list, tuple, Iterable]     # "SET" - DONT USE!
    ITERPATH_KEY = Union[Any, int]   # Any is for dict
    ITERPATH = tuple[ITERPATH_KEY, ...]

    # -----------------------------------------------------------------------------------------------------------------
    BOOL_DRAFT = Union[
        None,
        Any,                                # fixme: hide? does it need? for results like []/{}/()/0/"" think KEEP! it mean you must know that its expecting boolComparing in further logic!
        bool,                               # as main idea! as already final generic
        Exception,
        Callable[..., bool | Any | NoReturn | Exception],   # as main idea! to get final generic
        VALUE_SPECIAL.NOVALUE
    ]

    # -----------------------------------------------------------------------------------------------------------------
    RESULT__NONE = None
    RESULT__BOOL = bool
    RESULT__EXX = Exception     # | type[Exception] - DONT USE typeExx! from any process Exception comes as instance!

    RESULT__ANY_EXX = Any | Exception
    RESULT__ANY_RAISE = Any | NoReturn

    RESULT__BOOL_NONE = bool | None
    RESULT__BOOL_RAISE = bool | NoReturn
    RESULT__BOOL_EXX = bool | Exception
    RESULT__RAISE_NONE = NoReturn | None

    RESULT__BOOL_RAISE_NONE = bool | NoReturn | None

    # -----------------------------------------------------------------------------------------------------------------
    EXPECTED = bool | type[Any | Exception] | Any

    # -----------------------------------------------------------------------------------------------------------------
    CALLABLE_DRAFT = Union[Any, type[Any], Callable[..., Any | NoReturn]]
    # CALLABLE_FINAL    # dont need final! Final - are all others!

    CALLABLE__NONE = Callable[..., None]
    CALLABLE__BOOL = Callable[..., bool]
    CALLABLE__EXX = Callable[..., Exception]

    CALLABLE__BOOL_NONE = Callable[..., bool | None]
    CALLABLE__BOOL_RAISE = Callable[..., bool | NoReturn]
    CALLABLE__BOOL_EXX = Callable[..., bool | Exception]
    CALLABLE__RAISE_NONE = Callable[..., NoReturn | None]  # not expecting any bool! intended/inportant only raising as inappropriate position!

    CALLABLE__BOOL_RAISE_NONE = Callable[..., bool | NoReturn | None]

    # -----------------------------------------------------------------------------------------------------------------
    VALID_VALIDATOR = Union[
        Any,    # generic final instance as expecting value - direct comparison OR comparison instance like Valid!
        # Type,   # Class as validator like Exception????? fixme
        type[Exception],  # direct comparison
        Callable[[Any, ...], bool | NoReturn]     # func with first param for validating source
    ]


# =====================================================================================================================



"""
Designed to print info about object (properties+methods results)

But why? if we can use debugger directly?
Reason:
1. to get and save standard text info,
it useful to keep this info for future quick eye sight without exact condition like other OS or device/devlist/configuration
2. in debugger we cant see result of methods!
try to see for example information from platform module! it have only methods and no one in object tree in debugger!
```python
import platform

obj = platform
print(platform.platform())
pass    # place debug point here
```
3. Useful if you wish to see info from remote SOURCE if connecting directly over ssh for example

FEATURES
    "print all properties/methods results",
    "show exceptions on methods/properties",
    "skip names by full/part names and use only by partnames",
    "separated collections by groups",


TODO:
    "add TIMEOUT (use start in thread!) for print! use timeout for GETATTR!!!",
    [
        "realise PRINT_DIFFS=CHANGE_state/COMPARE_objects (one from different states like thread before and after start)!",
        "this is about to save object STATE!",
        "add parameter show only diffs or show all",
        "add TESTS after this step!",
    ],
    "apply asyncio.run for coroutine?",
    "merge items Property/Meth? - cause it does not matter callable or not (just add type info block)",
    "add check__instance_of_user_class",
"""
from typing import *

from base_aux.base_types.m0_static_aux import *
from base_aux.base_types.m1_type_aux import TypeAux
from base_aux.aux_attr.m0_static import check_name__buildin, NAMES__SKIP_PARTS


# =====================================================================================================================
def _value_search_by_list(source: Any, search_list: list[Any]) -> Any | None:
    search_list = search_list or []
    for search_item in search_list:
        try:
            if search_item in source:
                return search_item
        except:
            pass

        try:
            if search_item == source:
                return search_item
        except:
            pass


# =====================================================================================================================
class ObjectInfo:
    """
    :ivar MAX_ITER_ITEMS: 0 or None if not limited!
    """
    # SETTINGS --------------------------------------------
    MAX_LINE_LEN: int = 100
    MAX_ITER_ITEMS: int = 5
    HIDE_BUILD_IN: bool = None
    LOG_ITER: bool = None

    NAMES__USE_ONLY_PARTS: list[str] = []
    NAMES__SKIP_FULL: list[str] = []
    NAMES__SKIP_PARTS: list[str] = NAMES__SKIP_PARTS

    # AUX --------------------------------------------------
    SOURCE: Any = None
    STATE: ObjectState = None
    STATE_OLD: ObjectState | None = None

    def __init__(
            self,
            source: Optional[Any] = None,

            max_line_len: Optional[int] = None,
            max_iter_items: Optional[int] = None,
            hide_build_in: Optional[bool] = None,
            log_iter: Optional[bool] = None,

            names__use_only_parts: Union[None, str, list[str]] = None,
            names__skip_full: Union[None, str, list[str]] = None,
            names__skip_parts: str | list[str] = None,
    ):
        # SETTINGS -----------------------------------------------------------
        # RAPAMS -----------------------
        if max_line_len is not None:
            self.MAX_LINE_LEN = max_line_len
        if max_iter_items is not None:
            self.MAX_ITER_ITEMS = max_iter_items
        if hide_build_in is not None:
            self.HIDE_BUILD_IN = hide_build_in
        if log_iter is not None:
            self.LOG_ITER = log_iter

        # LISTS -----------------------
        if names__use_only_parts:
            if isinstance(names__use_only_parts, str):
                names__use_only_parts = [names__use_only_parts, ]
            self.NAMES__USE_ONLY_PARTS = names__use_only_parts
        if names__skip_full:
            if isinstance(names__skip_full, str):
                names__skip_full = [names__skip_full, ]
            self.NAMES__SKIP_FULL.extend(names__skip_full)
        if names__skip_parts:
            if isinstance(names__skip_parts, str):
                names__skip_parts = [names__skip_parts, ]
            self.NAMES__SKIP_PARTS = [*self.NAMES__SKIP_PARTS, *names__skip_parts]

        # WORK -----------------------------------------------------------
        self.state_reload()

        if source is not None:
            self.SOURCE = source

    # =================================================================================================================
    def state_clear(self) -> None:
        self.STATE_OLD = self.STATE
        self.STATE = ObjectState()

    def state_reload(self) -> None:
        self.state_clear()

        # WORK --------------------------------------
        group_name = f"log_iter={self.LOG_ITER}(wait last touched)"
        self._print_line__group_separator(group_name)

        for pos, name in enumerate(dir(self.SOURCE), start=1):
            if self.LOG_ITER:
                print(f"{pos}:\t{name}")

            # FILTER -----------------------------------------------------------------------
            if self.HIDE_BUILD_IN and check_name__buildin(name):
                continue

            if self.NAMES__USE_ONLY_PARTS:
                use_name = False
                for name_include_item in self.NAMES__USE_ONLY_PARTS:
                    if name_include_item.lower() in name.lower():
                        use_name = True
                        break
                if not use_name:
                    continue

            # SKIP ------------------------------------------------------------------------
            SKIP_FULLNAMES = []
            if self.NAMES__SKIP_FULL:
                SKIP_FULLNAMES.extend(self.NAMES__SKIP_FULL)
            if SKIP_FULLNAMES:
                SKIP_FULLNAMES.extend(SKIP_FULLNAMES)
            if name in SKIP_FULLNAMES:
                self.STATE.SKIPPED_FULLNAMES.append(name)
                continue

            SKIP_PARTNAMES = []
            if self.NAMES__SKIP_PARTS:
                SKIP_PARTNAMES.extend(self.NAMES__SKIP_PARTS)
            if SKIP_PARTNAMES:
                SKIP_PARTNAMES.extend(SKIP_PARTNAMES)
            if _value_search_by_list(source=name, search_list=SKIP_PARTNAMES):
                self.STATE.SKIPPED_PARTNAMES.append(name)
                continue

            # PROPERTIES/METHODS + Exception--------------------------------------------------
            attr_is_method: bool = False
            try:
                value = getattr(self.SOURCE, name)
            except Exception as exx:
                self.STATE.PROPERTIES__EXX.update({name: exx})
                continue

            if callable(value):
                attr_is_method = True
                try:
                    value = value()
                except Exception as exx:
                    self.STATE.METHODS__EXX.update({name: exx})
                    continue

            # print(f"{name=}/{attr_obj=}/type={type(attr_obj)}/elementary={isinstance(attr_obj, TYPES.ELEMENTARY)}")

            # PLACE VALUE ---------------------------------------------------------------------
            if TypeAux(value).check__elementary_single():
                if attr_is_method:
                    self.STATE.METHODS__ELEMENTARY_SINGLE.update({name: value})
                else:
                    self.STATE.PROPERTIES__ELEMENTARY_SINGLE.update({name: value})

            elif TypeAux(value).check__elementary_collection():
                if attr_is_method:
                    self.STATE.METHODS__ELEMENTARY_COLLECTION.update({name: value})
                else:
                    self.STATE.PROPERTIES__ELEMENTARY_COLLECTION.update({name: value})

            else:
                if attr_is_method:
                    self.STATE.METHODS__OBJECTS.update({name: value})
                else:
                    self.STATE.PROPERTIES__OBJECTS.update({name: value})

    # def state_diff_last(self):
    #     # NOTE: use attrsDUMP!!! instead! cause raised values may be moved into other group! (single/collection)
    #     self.state_reload()
    #     result = ObjectState()
    #     for (name_g1, values_g1), (name_g2, values_g2) in zip(self.STATE_OLD.items(), self.STATE.items()):
    #         for

    # =================================================================================================================
    def _print_line__group_separator(self, group_name: str) -> str:
        """
        GOAL MAIN - print!
        GOAL SECONDARY - return str - just for tests!!!
        """
        result = "-" * 10 + f"{group_name:-<90}"      # here is standard MAX_LINE_LEN
        print(result)
        return result

    def _print_line__name_type_value(self, name: Optional[str] = None, type_replace: Optional[str] = None, value: Union[None, Any, ItemKeyValue] = None, intend: Optional[int] = None) -> str:
        # -------------------------------
        name = name or ""
        if isinstance(value, ItemKeyValue):
            name = ""
        block_name = f"{name}"

        # -------------------------------
        block_type = f"{value.__class__.__name__}"
        if isinstance(value, ItemKeyValue):
            block_type = f"{value.KEY.__class__.__name__}:{value.VALUE.__class__.__name__}"
        if type_replace is not None:
            block_type = type_replace

        # -------------------------------
        intend = intend or 0
        if isinstance(value, ItemKeyValue):
            intend = 1

        _block_intend = "\t" * intend

        # -------------------------------
        block_value = f"{value}"
        if isinstance(value, ItemKeyValue):
            block_type = f"{value.KEY}:{value.VALUE}"
        elif TypeAux(value).check__exception():
            block_value = f"{value!r}"

        # -------------------------------
        result = f"{block_name:20}\t{block_type:12}:{_block_intend}{block_value}"

        if self.MAX_LINE_LEN and len(result) > self.MAX_LINE_LEN:
            result = result[:self.MAX_LINE_LEN - 3*2] + "..."

        # --------------------------------------------------------------------------------------
        print(result)

        # -------------------------------
        if name and str(value) != repr(value) and str(value) != str(block_value) and not TypeAux(value).check__exception():
            # additional print repr()
            self._print_line__name_type_value(name=None, type_replace="__repr()", value=repr(value))

        return result

    # =================================================================================================================
    def _print_block__head(self) -> None:
        # start printing ----------------------------------
        group_name = f"{self.__class__.__name__}.print"
        self._print_line__group_separator(group_name.upper())

        print(f"str(SOURCE)={str(self.SOURCE)}")
        print(f"repr(SOURCE)={repr(self.SOURCE)}")
        print(f"type(SOURCE)={type(self.SOURCE)}")

        try:
            mro = self.SOURCE.__class__.__mro__
        except:
            mro = self.SOURCE.__mro__

        mro = [cls.__name__ for cls in mro]

        print(f"mro(SOURCE)={mro}")

        # SETTINGS ----------------------------------------
        group_name = "SETTINGS"
        self._print_line__group_separator(group_name)

        print(f"{self.NAMES__USE_ONLY_PARTS=}")
        print(f"{self.NAMES__SKIP_FULL=}")
        print(f"{self.NAMES__SKIP_PARTS=}")

        print(f"{self.HIDE_BUILD_IN=}")
        print(f"{self.LOG_ITER=}")

        print(f"{self.MAX_LINE_LEN=}")
        print(f"{self.MAX_ITER_ITEMS=}")

    def _print_block__name_value(self, name, value) -> None:
        # ALWAYS ---------------------------------------------------------------------------------
        self._print_line__name_type_value(name=name, value=value)
        if len(str(value)) <= self.MAX_LINE_LEN:
            return

        # COLLECTION -----------------------------------------------------------------------------
        if TypeAux(value).check__elementary_collection():
            # start some pretty style -------------------------------------
            if not isinstance(value, dict):
                _index = 0
                for item in value:
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=item, intend=1)

            elif isinstance(value, dict):
                _index = 0
                for item_key, item_value in value.items():
                    _index += 1
                    if self.MAX_ITER_ITEMS and _index > self.MAX_ITER_ITEMS:
                        self._print_line__name_type_value(name=None, type_replace="", value="...", intend=1)
                        break
                    self._print_line__name_type_value(name=None, value=ItemKeyValue(item_key, item_value))

        # SINGLE/EXX/OBJECTS ---------------------------------------------------------------------
        if any([
            TypeAux(value).check__elementary_single(),
            TypeAux(value).check__exception(),
            TypeAux(value).check__instance(),
        ]):
            pass    # DONT USE RETURN HERE OR ELIF IN NEXT LINE!!!

    # =================================================================================================================
    def print(self) -> None:
        """print all params from object
        if callable - try to call it!
        """
        WRAPPER_MAIN_LINE = "="*min(90, self.MAX_LINE_LEN)  # here we need use less then

        print(WRAPPER_MAIN_LINE)
        self._print_block__head()

        for group_name, group_values in self.STATE.__getstate__().items():
            self._print_line__group_separator(group_name)

            if TypeAux(group_values).check__elementary_collection_not_dict():
                for pos, name in enumerate(group_values, start=1):
                    print(f"{pos}:\t{name}")
            else:
                for name, value in group_values.items():
                    self._print_block__name_value(name, value)

        print(WRAPPER_MAIN_LINE)

    # =================================================================================================================
    @classmethod
    def print_diffs(cls, *state: ObjectState) -> None:
        pass
        # TODO: FINISH!


# =====================================================================================================================



from typing import *
from pathlib import Path

from base_aux.testplans.devices import *
from base_aux.aux_attr.m4_kits import *
from base_aux.path2_file.m4_fileattrs import *
from base_aux.aux_datetime.m1_datetime import *


# =====================================================================================================================
class Base_Stand:
    NAME: str = "[DEF] STAND NAME"
    DESCRIPTION: str = "[DEF] STAND DESCRIPTION"
    SN: str = "[DEF] STAND SN"

    DIRPATH_RESULTS: Union[str, Path] = "RESULTS"

    DEV_LINES: DeviceKit

    # TCSc_LINE: dict[type, bool]   # TODO: use TableLine??? - NO! KEEP DICT! with value like USING! so we can use one
    TCSc_LINE: TableLine = TableLine()

    TIMESTAMP_START: DateTimeAux | None = None
    TIMESTAMP_STOP: DateTimeAux | None = None

    # =================================================================================================================
    def __init__(self) -> None:
        # PREPARE CLSs ========================================
        for tc_cls in self.TCSc_LINE:
            # init STAND -----------------------------------
            tc_cls.STAND = self

            # gen INSTS -----------------------------------
            tcs_insts = []
            for index in range(self.DEV_LINES.COUNT_COLUMNS):
                tc_i = tc_cls(index=index)
                tcs_insts.append(tc_i)
            tc_cls.TCSi_LINE = TableLine(*tcs_insts)    # TODO: move into TC_CLS

    # =================================================================================================================
    def stand__get_info__general(self) -> dict[str, Any]:
        result = {
            "STAND.NAME": self.NAME,
            "STAND.DESCRIPTION": self.DESCRIPTION,
            "STAND.SN": self.SN,

            "STAND.TIMESTAMP_START": str(self.TIMESTAMP_START),
            "STAND.TIMESTAMP_STOP": str(self.TIMESTAMP_STOP),
        }
        return result

    def stand__get_info__tcs(self) -> dict[str, Any]:
        """
        get info/structure about stand/TP
        """
        TP_TCS = []
        for tc_cls in self.TCSc_LINE:
            TP_TCS.append(tc_cls.tcc__get_info())

        result = {
            "TESTCASES": TP_TCS,
            # "TP_DUTS": [],      # TODO: decide how to use
            # [
            #     # [{DUT1}, {DUT2}, …]
            #     {
            #         DUT_ID: 1  # ??? 	# aux
            #         DUT_SKIP: False
            #     }
            # ]

            }
        return result

    def stand__get_info__general_tcsc(self) -> dict[str, Any]:
        """
        get info/structure about stand/TP
        """
        result = {
            **self.stand__get_info__general(),
            **self.stand__get_info__tcs(),
        }
        return result

    # -----------------------------------------------------------------------------------------------------------------
    def stand__get_results(self) -> dict[str, Any]:
        """
        get all results for stand/TP
        """
        TCS_RESULTS = {}
        for tc_cls in self.TCSc_LINE:
            TCS_RESULTS.update({tc_cls: tc_cls.tcsi__get_results()})

        result = {
            "STAND" : self.stand__get_info__general(),
            "TCS": TCS_RESULTS,
        }
        return result

    def stand__save_results(self) -> None:
        for index in range(self.DEV_LINES.COUNT_COLUMNS):
            result_i_short = {}
            result_i_full = {}
            for tc_cls in self.TCSc_LINE:
                tc_inst = None
                try:
                    tc_inst: 'Base_TestCase' = tc_cls.TCSi_LINE[index]

                    tc_inst_result_full = tc_inst.tci__get_result(add_info_dut=False, add_info_tc=False)
                    tc_inst_result_short = tc_inst_result_full["tc_result"]
                except:
                    tc_inst_result_short = None
                    tc_inst_result_full = None

                result_i_short.update({tc_cls.DESCRIPTION: tc_inst_result_short})
                result_i_full.update({tc_cls.DESCRIPTION: tc_inst_result_full})

            DUT = tc_inst.DEV_COLUMN.DUT

            if not DUT.DEV_FOUND or not DUT.DUT_FW:
                continue

            dut_info = DUT.dev__get_info()
            result_dut = {
                "STAND": self.stand__get_info__general(),
                "DUT": dut_info,
                "RESULTS_SHORT": result_i_short,
                "RESULTS_FULL": result_i_full,
            }

            # data_text = json.dumps(result_dut, indent=4, ensure_ascii=False)

            filename = f"{self.TIMESTAMP_STOP}[{index}].json"
            filepath = pathlib.Path(self.DIRPATH_RESULTS, filename)

            tfile = TextFile(text=str(result_dut), filepath=filepath)
            tfile.pretty__json()
            tfile.write__text()


# =====================================================================================================================
# if __name__ == "__main__":
#     print(load__tcs())


# =====================================================================================================================
